/*
FILE:     CifParserBase.C
*/
/*
VERSION:  6.302
*/
/*
DATE:     4/26/2005
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2005 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    A DDL 2.1 compliant CIF file parser.
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define  MAX_CIF_NAME_LENGTH  100

#include "CifScannerBase.h"
#include "CifParserInt.h"
#include "CifFileObj.h"
#include "CifParserBase.h"
#include "CifString.h"

class CifFileObj;
class CifFileObjBase;

extern FILE* cifparser_in;
extern int cifparser_parse();

char Glob_tBufKeyword[MAX_CIF_NAME_LENGTH];
char* Glob_pBufValue;
CifString* Glob_sBufValue;
char* Glob_dataBlockName;

CifParser* CifParserP = NULL;

CifParser::CifParser(FILE* in, CifFileObj *fo, int verbose, CifFileReadDef readDef)
{
    cifparser_in = in;
    _verbose=verbose;
    Clear();
    errorLog.Clear();
    if (fo)
        _fobj = fo;
    else
        exit(-1);

    OpenLog("CifParser.log",_verbose);
    _readDef = readDef;

    CifParserP = this;
}

void CifParser::Parse()
{
    int ret;

    ret = cifparser_parse();
    if (ret != 0)
    {
        int b = 0;
        b++;
    }
}

void cifparser_error(const char *s)
{
    CifParserP->Error(s);
}

void CifParser::Error(const char* s)
/* 
 * Purpose:  yyerror() Print errors in CifParsr.log.
 */
{
  errorLog += "ERROR - ";
  errorLog += s;
  errorLog += " at line ";
  errorLog += NDBlineNo;
#if DEBUG
  errorLog += "   ";
  errorLog += yytext;
#endif
  errorLog += '\n';
  log <<"ERROR - "<< s << " at line " <<  NDBlineNo << endl;
  _err++;
}

void CifParser::Reset() 
{
  if (_tbl && (_curItemNo > 0)) { // write the current table / management of _tbl by _fobj
      _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
      _nTablesInBlock++;
  }
  if (_rowBuf) delete _rowBuf;
  if (_fieldList) free(_fieldList);
}

void CifParser::Clear()
{
  _nTablesInBlock=0;
  _numDataBlocks=0;
  _afterLoop = FALSE;
  _curItemNo=0;
  _curRow=0;
  _curCol=0;
  _curValueNo=0;
  _itemIndex=0;
  _fieldListAlloc=100;
  _fieldList=(int *) malloc(_fieldListAlloc*sizeof(int));
  _pBufValue=NULL;
  _sBufValue=NULL;
  _pBufKeyword=NULL;
  _rowBuf= new ReVarCifArray<CifString>;
  _fobj=NULL;
  _tbl=NULL;
  _err=0;
  _warn=0;
  _uString.Copy(".");
  _mString.Copy("?");
  strcpy(_tBufKeyword,"");
  strcpy(_curCategoryName,"");
  strcpy(_curDataBlockName,"UNNAMED-0");  
  strcpy(_prevDataBlockName,"UNNAMED-0");
}




void CifParser::ProcessLoopDeclaration(void)
/* ----------------------------------------------------------------------
     Purpose: CifParser::ProcessLoopDeclaration(void)

              Handles initialization for a new loop, by creating a new 
              category and adding the current item name to this category.
   ---------------------------------------------------------------------- */
{
  char  categoryName[MAX_CIF_NAME_LENGTH];

#if DEBUG
  if (_verbose) log << "Processing loop declaration at line " << NDBlineNo << " value " <<_tBufKeyword << endl;
#endif
  

  _afterLoop=TRUE;
  if (! CifFileObjBase::GetCategoryNameFromItemName(categoryName, _tBufKeyword) ) {
    log << "Warning -  Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Warning - Error in category name at line ";
    errorLog += NDBlineNo;
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _warn++;
  }

  if (_readDef.Category_OK(categoryName)&&_readDef.Datablock_OK(_curDataBlockName)) {
  if (_fobj->IsTablePresent(categoryName)) {
    log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
    errorLog += "Warning - Duplicate category name ";
    errorLog += categoryName;
    errorLog += " at line ";
    errorLog += NDBlineNo;
    errorLog += '\n';
    _warn++;
  } else {
    if (_tbl) { // write the current table / management of _tbl by _fobj
      _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
      _nTablesInBlock++;
    }
    _tbl = new ISTable(categoryName);
    _curRow = 0;
    strcpy(_curCategoryName,categoryName);
    ProcessItemNameList();
  }
  }
  else
  {
    if (_fobj->IsTablePresent(categoryName)) {
      log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
      errorLog += "Warning - Duplicate category name ";
      errorLog += categoryName;
      errorLog += " at line ";
      errorLog += NDBlineNo;
      errorLog += '\n';
      _warn++;
    } else {
    if (_tbl) { // write the current table / management of _tbl by _fobj
      _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
      _nTablesInBlock++;
  
      _tbl = NULL;
      }
    }
    strcpy(_curCategoryName,categoryName);
  }
}


void CifParser::ProcessItemNameList(void)
/* ----------------------------------------------------------------------
   Purpose: CifParser::ProcessItemNameList(void)

            Registers the item keyword for the the current item in the 
	    current category.  Maintains an index array of "valid" keyword 
	    names in fieldList[].  This array is used to indirectly 
	    reference between keywords and values ...  
 * ----------------------------------------------------------------------*/
{

  char keywordName[MAX_CIF_NAME_LENGTH], categoryName[MAX_CIF_NAME_LENGTH];

#if DEBUG
  if (_verbose) log << "Processing item name list at line " <<  NDBlineNo << " keyword " <<  _tBufKeyword << endl;
#endif

  if (! CifFileObjBase::GetCategoryNameFromItemName(categoryName, _tBufKeyword) ) {
    log << "Warning - Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Warning - Error in category name at line ";
    errorLog += NDBlineNo;
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _warn++;
  }

  if (_readDef.Category_OK(categoryName)&&_readDef.Datablock_OK(_curDataBlockName)) {
  if ( _curItemNo > _fieldListAlloc - 1) {
    _fieldListAlloc = _curItemNo + _fieldListAlloc; 
    _fieldList = (int *) realloc(_fieldList, (_fieldListAlloc) * sizeof(int));
  }


  if ( !strcmp(categoryName,_curCategoryName) && CifFileObjBase::GetKeywordFromItemName(keywordName, _tBufKeyword)) {
    if (_tbl->GetColumnIndex(keywordName) < 0) {
#if DEBUG
    if (_verbose) log << "Line " << NDBlineNo << " keyword is " << keywordName << endl;
#endif
    _tbl->AddColumn(keywordName);
    _itemIndex++;
    _fieldList[_curItemNo] = _itemIndex;
    }
    else { 
      log << "Warning - Duplicate item name " <<_tBufKeyword  << " at line " << NDBlineNo << endl;
      errorLog += "Warning - Duplicate item name ";
      errorLog += _tBufKeyword;
      errorLog += " at line ";
      errorLog += NDBlineNo;
      errorLog += '\n';
      _warn++;
      return;
    }
  } else {
    log << "Warning - Syntax error at line " << NDBlineNo << " at item " << _tBufKeyword << endl;
    errorLog += "Warning - Syntax error at line ";
    errorLog += NDBlineNo;
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _fieldList[_curItemNo] = -1;
    _warn++;
  }
  _curItemNo++;
  }

}
void CifParser::ProcessValueList(void)
/* ----------------------------------------------------------------------
     Purpose:  CifParser::ProcessValueList(void)

               Add the current value to the appropriate column in the 
               the current row.  Start a new row if necessary.
 * ----------------------------------------------------------------------*/
{
  CifString cs;
  if (_readDef.Category_OK(_curCategoryName)&&_readDef.Datablock_OK(_curDataBlockName)) {
#if DEBUG
  if (_verbose) {
    if (_pBufValue) 
      log << "Processing value at line " << NDBlineNo << " value [" << _pBufValue << "]" << endl;
    else 
      log << "Processing value at line " << NDBlineNo << " value [" << (*_sBufValue).Text() << "]" << endl;
  }
#endif

  if (_fieldList[_curValueNo] > 0 ) {
    
    if (_fieldList[_curValueNo] == 1) {
      _rowBuf->Clear();
      _curRow++;
       _tbl->AddRow();
		 _tbl->FillRow(*_rowBuf,_curRow-1);
    }

    if (_pBufValue) {
      if (strlen(_pBufValue) > 0) {
		  cs.Clear(); cs.Copy(_pBufValue);
	 _tbl->UpdateCell(cs,_curValueNo , _curRow-1);
      } else {
	 _tbl->UpdateCell(_mString,_curValueNo , _curRow-1);
      }
    } else {
      if (_sBufValue->Length() > 0)
	 _tbl->UpdateCell(*_sBufValue,_curValueNo , _curRow-1);
     else
	 _tbl->UpdateCell(_mString,_curValueNo , _curRow-1);
    }
  }
  _curValueNo++;

  if (_curValueNo == _curItemNo) {
#if DEBUG
  if (_verbose) {
    log << "Loading row " << _curRow -1 << " with " <<  _rowBuf->Length() << " elements" << endl;
    for (int i=0; i < _rowBuf->Length(); i++) {
      log << "Column [" << i << "] value "<< (*_rowBuf)[i].Text() << endl;
    }
  }
#endif
    _curValueNo = 0;
  }
  }
}





void CifParser::ProcessItemValuePair(void)
/* ----------------------------------------------------------------------
      Purpose: CifParser::ProcessItemValuePair()

               Assign the current value to its associated item name.
 * ----------------------------------------------------------------------*/
{
  char categoryName[MAX_CIF_NAME_LENGTH], keywordName[MAX_CIF_NAME_LENGTH];
  CifString cs;

  _curItemNo  = 1;
  _curValueNo = 0;
#if DEBUG
  if (_verbose)  {
    if (_pBufValue) 
      log << "Processing " << _tBufKeyword << " at " <<  NDBlineNo << " value " << _pBufValue  << endl;
    else 
      log << "Processing " << _tBufKeyword << " at " <<  NDBlineNo << " value " << (*_sBufValue).Text()  << endl;
  }
#endif

  if (! CifFileObjBase::GetCategoryNameFromItemName(categoryName, _tBufKeyword) ) {
    log << "Warning - Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Warning - Error in category name at line ";
    errorLog += NDBlineNo;
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _warn++;
    return;
  }
  if (_readDef.Category_OK(categoryName)&&_readDef.Datablock_OK(_curDataBlockName)) {

  if (_fobj->IsTablePresent(categoryName)) { //  duplicates a persistent table?
    log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
    errorLog += "Warning - Duplicate category name ";
    errorLog += categoryName;
    errorLog += " at line ";
    errorLog += NDBlineNo;
    errorLog += '\n';
    _warn++;
    return;
  } 

  if (strcmp(categoryName,_curCategoryName)|| _afterLoop) { 
    if (_tbl) { // write the current table / management of _tbl by _fobj
      _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
      _nTablesInBlock++;
    }      

   if (!strcmp(categoryName,_curCategoryName) && _afterLoop) { //  duplicates a persistent table-immediately after 
                                                            // loop goes pair item-value for same category
    log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
    errorLog += "Warning - Duplicate category name ";
    errorLog += categoryName;
    errorLog += " at line ";
    errorLog += NDBlineNo;
    errorLog += '\n';
    _warn++;
    return;
  } 
    _tbl = new ISTable(categoryName);
    _curCol = 0;   
    strcpy(_curCategoryName,categoryName);
  }
  _afterLoop=FALSE;
  if ( CifFileObjBase::GetKeywordFromItemName(keywordName, _tBufKeyword) ) {
    if (_tbl->GetColumnIndex(keywordName) < 0) {
#if DEBUG
    if (_verbose) log << "Line " << NDBlineNo << " keyword is " << _tBufKeyword << endl;
#endif
    _tbl->AddColumn(keywordName);
    _curCol++;
  }
    else { 
      log << "Warning - Duplicate item name " <<_tBufKeyword  << " at line " << NDBlineNo << endl;
      errorLog += "Warning - Duplicate item name ";
      errorLog += _tBufKeyword;
      errorLog += " at line ";
      errorLog += NDBlineNo;
      errorLog += '\n';
      _warn++;
      return;
    }
  }
  else {
    log << "Warning - Syntax error line at " << NDBlineNo << " item " << _tBufKeyword << endl;
    errorLog += "Warning - Syntax error at line ";
    errorLog += NDBlineNo;
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _warn++;
  }
  if (_pBufValue) {
    cs.Clear(); 
    if (strlen(_pBufValue) > 0) {
      cs.Copy(_pBufValue);
      _tbl->AddElementToColumn( cs, _curCol-1);
    } else
      _tbl->AddElementToColumn( _mString, _curCol-1);

  } else {
      if (_sBufValue->Length() > 0)
	_tbl->AddElementToColumn( *_sBufValue, _curCol-1);
      else
	_tbl->AddElementToColumn( _mString, _curCol-1);
  }
  }  
  else
  {
    if (_fobj->IsTablePresent(categoryName)) { //  duplicates a persistent table?
      log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
      errorLog += "Warning - Duplicate category name ";
      errorLog += categoryName;
      errorLog += " at line ";
      errorLog += NDBlineNo;
      errorLog += '\n';
      _warn++;
    }
    else
    {
      if (_tbl) { // write the current table / management of _tbl by _fobj
        _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
        _nTablesInBlock++;

        _tbl = NULL;
      }
    }
    strcpy(_curCategoryName,categoryName);
  }
}


void CifParser::ProcessAssignments()
{
  if (_curValueNo!=0) {
  log <<"Warning - Number of data values is not exact multiples of the number of data names   (look above line "<<NDBlineNo<<")"<<endl;
  errorLog += "Warning - Number of data values is not exact multiples of the number of data names   (look above line ";
  errorLog += NDBlineNo;
  errorLog += ")";
  errorLog += '\n';
  _warn++;
  }
}

void CifParser::ProcessLoop()
{
    _curItemNo = 0;
    _curValueNo = 0;
    _itemIndex = 0;
}

void CifParser::ProcessItemName()
{
    strcpy(_tBufKeyword, Glob_tBufKeyword);
}

void CifParser::ProcessItemValue()
{
    _pBufValue = Glob_pBufValue;
    _sBufValue = NULL;
}

void CifParser::ProcessLsItemValue()
{
    _pBufValue = NULL;
    _sBufValue = Glob_sBufValue;
}

void CifParser::ProcessUnknownValue()
{
   _sBufValue = &_uString;
   _pBufValue = NULL;
}

void CifParser::ProcessMissingValue()
{
   _sBufValue = &_mString;
   _pBufValue = NULL;
}

void CifParser::ProcessDataBlockName()
{

  // Save any unsaved data before changing state...
  //
  if (_tbl && (_curItemNo > 0) && _curDataBlockName && _curCategoryName) {
       if (_verbose) log << " Save category " << _curCategoryName << " in " << _curDataBlockName << endl;
      _fobj->WriteTable(_tbl,_curDataBlockName,_curCategoryName);
      _nTablesInBlock++;
      _tbl=NULL;
  }

  strcpy(_prevDataBlockName,_curDataBlockName);

  _numDataBlocks++;
  _nTablesInBlock=0;

  // Set current block name   - assign dummy names to unnamed blocks...
  if (&(Glob_dataBlockName)[5] && (strlen(&(Glob_dataBlockName)[5])>0)) {
    _fobj->SetDataBlock(&(Glob_dataBlockName)[5]);
    strcpy(_curDataBlockName,&(Glob_dataBlockName)[5]);
  } else {
    char bCount[10];
    sprintf(bCount,"%d",_numDataBlocks);
    strcpy(_curDataBlockName,"UNNAMED-");
    strcat(_curDataBlockName,bCount);
  }

  strcpy(_curCategoryName,"");

#if DEBUG
  if (_verbose) log << " Previous data block is now  " <<  _prevDataBlockName << endl;
  if (_verbose) log << " Current  data block is now  " <<  _curDataBlockName << endl;
#endif
}

CifParser::~CifParser()
{
    Reset();
    CifParserP = NULL;
}

void ProcessAssignmentsFromParser()
{
    CifParserP->ProcessAssignments();
}

void ProcessLoopFromParser()
{
    CifParserP->ProcessLoop();
}

void ProcessItemValuePairFromParser()
{
    CifParserP->ProcessItemValuePair();
}

void ProcessLoopDeclarationFromParser()
{
    CifParserP->ProcessLoopDeclaration();
}

void ProcessItemNameListFromParser()
{
    CifParserP->ProcessItemNameList();
}

void ProcessValueListFromParser()
{
    CifParserP->ProcessValueList();
}

void ProcessItemNameFromParser()
{
    CifParserP->ProcessItemName();
}

void ProcessItemValueFromParser()
{
    CifParserP->ProcessItemValue();
}

void ProcessLsItemValueFromParser()
{
    CifParserP->ProcessLsItemValue();
}

void ProcessUnknownValueFromParser()
{
    CifParserP->ProcessUnknownValue();
}

void ProcessMissingValueFromParser()
{
    CifParserP->ProcessMissingValue();
}

void ProcessDataBlockNameFromParser()
{
    CifParserP->ProcessDataBlockName();
}

